/*
Copyright (c) 2007-2013 Alloysoft, LLC. All rights reserved.
http://www.alloysoft.com
*/

MediaItem = function(t) {
	if (t === undefined) t = {};
	this.mediaId = t.mediaId ? t.mediaId : null;
	this.container = t.container ? t.container : false;
	this.artist = t.artist ? t.artist : "";
	this.album = t.album ? t.album : "";
	this.title = t.title ? t.title : "";
	this.rating = t.rating ? t.rating : 0;
	this.length = t.length ? t.length : 0;
};

MediaItem.prototype.equals = function(o) {
	for (var p in this) {
		if (this.hasOwnProperty(p)) {
			if (!o.hasOwnProperty(p)) {
				return false;
			}

			if (this[p] !== o[p]) {
				return false;
			}
		}
	}

	return true;
}

var PlayerError = {
	UIScriptingDisabled: 106
};

// Define JSON.parse if it doesn't exist
var JSON;
if (!JSON) {
	JSON = {};
}

if (typeof JSON.parse !== 'function') {
	JSON.parse = function(text) {
		return eval('(' + text + ')');
	};
}

var player = new function() {
	// Public properties /////////////////////////////////////////////////

	this.currentMedia = new MediaItem();
	this.volume = 0;
	this.playingState = "stopped";
	this.playbackPosition = 0;
	this.playlistPosition = 0;
	this.shuffleState = false;
	this.repeatState = false;
	this.currentArtURL = "";
	this.speakers = [];
	this.libraryVersion = "";

	this.volumeIncrement = 10;
	this.albumArtSize = 320;
	this.navigatorOnlineEnabled = true;

	// event handlers
	this.onConnectedToPlayer = null;
	this.onCurrentMediaChanged = null;
	this.onAlbumArtChanged = null;
	this.onRatingChanged = null;
	this.onVolumeChanged = null;
	this.onPlayingStateChanged = null;
	this.onShuffleStateChanged = null;
	this.onRepeatStateChanged = null;
	this.onPlaylistChanged = null;
	this.onPlaylistPositionChanged = null;
	this.onPlaybackPositionChanged = null;
	this.onSpeakerListChanged = null;
	this.onResponseError = null;
	this.onMediaLibraryChanged = null;
	this.onBrowseLibraryResponse = null;
	this.onSearchLibraryResponse = null;
	this.onGetPlaylistItemsResponse = null;

	// Private variables /////////////////////////////////////////////////

	var self = this;
	var debug = false;
	var serverURL = '';

	var http_request;
	var currentRequest = '';
	var isRequestPending = false;
	var pendingRequests = [];

	var sleepMonitorTime = new Date();
	var linkRecoveryActive = false;
	var linkRecoveryTimeout = undefined;
	var lastStateRev;

	// Privledged methods ////////////////////////////////////////////////

	this.start = function(host) {
		// use the origin by default, allow specification of a host for embedding
		serverURL = (host === undefined) ? '' : 'http://' + host + ':3569';

		if (this.navigatorOnlineEnabled && navigator.onLine !== undefined) {
			if (document.body.addEventListener) {
				document.body.addEventListener('online', function() {
					getState();
				}, false);
			}
		} else {
			// use the sleep monitor if offline/online events are not available
			sleepMonitorTime = new Date();
			setInterval(sleepMonitor, 1000);
		}

		requestInitialState();
	};

	this.playPause = function() {
		sendRequest('<PlayPauseRequest></PlayPauseRequest>');
	};

	this.play = function() {
		sendRequest('<PlayRequest></PlayRequest>');
	};

	this.pause = function() {
		sendRequest('<PauseRequest></PauseRequest>');
	};

	this.stop = function() {
		sendRequest('<StopRequest></StopRequest>');
	};

	this.previous = function() {
		sendRequest('<PreviousItemRequest></PreviousItemRequest>');
	};

	this.next = function() {
		sendRequest('<NextItemRequest></NextItemRequest>');
	};

	this.toggleShuffleState = function()  {
		sendRequest('<SetShuffleStateRequest><shuffleState>' + !this.shuffleState + '</shuffleState></SetShuffleStateRequest>');
	};

	this.toggleRepeatState = function()  {
		sendRequest('<SetRepeatStateRequest><repeatState>' + !this.repeatState + '</repeatState></SetRepeatStateRequest>');
	};

	this.setVolume = function(volume)  {
		if (volume < 0) {
			volume = 0;
		} else if (volume > 100) {
			volume = 100;
		}

		this.volume = volume;
		sendRequest('<SetVolumeRequest><volume>' + volume + '</volume></SetVolumeRequest>', 'setvolume');
	};

	this.setRating = function(rating) {
		if (rating < 0) {
			rating = 0;
		} else if (rating > 100) {
			rating = 100;
		}

		sendRequest('<SetTrackRatingRequest><rating>' + rating + '</rating></SetTrackRatingRequest>');
	};

	this.volumeUp = function() {
		this.setVolume(this.volume + this.volumeIncrement);

		if (onVolumeChanged) {
			onVolumeChanged(this.volume);
		}
	};

	this.volumeDown = function() {
		this.setVolume(this.volume - this.volumeIncrement);

		if (onVolumeChanged) {
			onVolumeChanged(this.volume);
		}
	};

	this.playItem = function(item, startingItem) {
		if (isNaN(Number(item))) {
			// a media library URL was specified
			if (startingItem !== undefined) {
				sendRequest('<PlayMediaRequest><libPath>' + item + '</libPath><startingItem>' + startingItem + '</startingItem></PlayMediaRequest>');
			} else {
				sendRequest('<PlayMediaRequest><libPath>' + item + '</libPath></PlayMediaRequest>');
			}
		} else {
			// a media library ID was specified
			sendRequest('<PlayMediaRequest><id>' + item + '</id></PlayMediaRequest>');
		}
	};

	this.playlistAdd = function(item) {
		if (isNaN(Number(item))) {
			// a media library URL was specified
			sendRequest('<PlaylistAddRequest><libPath>' + item + '</libPath></PlaylistAddRequest>');
		} else {
			// a media library ID was specified
			sendRequest('<PlaylistAddRequest><id>' + item + '</id></PlaylistAddRequest>');
		}
	};

	this.setPlaylistPosition = function(index) {
		sendRequest('<SetPlaylistPositionRequest><position>' + index + '</position></SetPlaylistPositionRequest>');
	};

	this.setPlaybackPosition = function(position) {
		sendRequest('<SetPlayPositionRequest><position>' + position + '</position></SetPlayPositionRequest>', 'setplaypos');
	};

	this.selectSpeaker = function(id) {
		sendRequest('<SelectSpeakerRequest><id>' + id + '</id></SelectSpeakerRequest>');
	};

	this.toggleSpeaker = function(id, enabled) {
		if (enabled === undefined) {
			for (var i=0; i < this.speakers.length; i++) {
				var speaker = this.speakers[i];
				if (speaker.id == id) {
					enabled = !speaker.enabled;
					break;
				}
			}
		}
		sendRequest('<EnableSpeakerRequest><id>' + id + '</id><enabled>' + enabled + '</enabled></EnableSpeakerRequest>');
	};

	this.getPlaylistItems = function(start, count, includeTotalCount, callback) {
		start = (start === undefined) ? 0 : start;
		count = (count === undefined) ? -1 : count;
		includeTotalCount = (includeTotalCount === undefined) ? true : includeTotalCount;
		callback = (callback === undefined) ? this.onGetPlaylistItemsResponse : callback;

		var url = '/playlist.json?start=' + start + '&count=' + count;

		if (includeTotalCount) {
			url += '&includeTotalCount=true';
		}

		sendRestRequest(url, callback);
	};

	this.browseLibrary = function(browsePath, start, count, includeTotalCount, callback) {
		start = (start === undefined) ? 0 : start;
		count = (count === undefined) ? -1 : count;
		includeTotalCount = (includeTotalCount === undefined) ? true : includeTotalCount;
		callback = (callback === undefined) ? this.onBrowseLibraryResponse : callback;

		// if browsing based on a start index (not a letter) validate the start position
		if (!isNaN(Number(start))) {
			if (start < 0) {
				start = 0;
			}
		}

		var url = '/media' + browsePath + '.json?start=' + start + '&count=' + count;

		if (includeTotalCount) {
			url += '&includeTotalCount=true';
		}

		sendRestRequest(url, callback);
	};

	this.searchLibrary = function(searchTerms, start, count, callback) {
		start = (start === undefined) ? 0 : start;
		count = (count === undefined) ? -1 : count;
		callback = (callback === undefined) ? this.onSearchLibraryResponse : callback;

		if (start < 0) {
			start = 0;
		}

		//sendRequest('<MediaSearchRequest><start>' + start + '</start><count>' + count + '</count><any>' + encodeURIComponent(searchTerms) + '</any></MediaSearchRequest>');

		var url = '/media/search/' + encodeURIComponent(searchTerms) + '.json?start=' + start + '&count=' + count;
		sendRestRequest(url, callback);
	};

	this.refreshState = function() {
		if (this.onCurrentMediaChanged) {
			this.onCurrentMediaChanged(this.currentMedia);
		}

		if (this.onAlbumArtChanged) {
			this.onAlbumArtChanged(this.currentArtURL);
		}

		if (this.onRatingChanged) {
			this.onRatingChanged(this.currentMedia.rating);
		}

		if (this.onVolumeChanged) {
			this.onVolumeChanged(this.volume);
		}

		if (this.onPlayingStateChanged) {
			this.onPlayingStateChanged(this.playingState);
		}

		if (this.onShuffleStateChanged) {
			this.onShuffleStateChanged(this.shuffleState);
		}

		if (this.onRepeatStateChanged) {
			this.onRepeatStateChanged(this.repeatState);
		}

		if (this.onPlaybackPositionChanged) {
			this.onPlaybackPositionChanged(this.playbackPosition);
		}
	};

	// Private methods ///////////////////////////////////////////////////

	function createHTTPRequest() {
		var xmlHttp = null;

		if (window.XMLHttpRequest) {
			xmlHttp = new XMLHttpRequest();
			if (xmlHttp.overrideMimeType) {
				xmlHttp.overrideMimeType('text/xml');
			}
		} else if (window.ActiveXObject) {
			try {
				xmlHttp = new ActiveXObject("Msxml2.XMLHTTP");
			} catch (e) {
				try {
					xmlHttp = new ActiveXObject("Microsoft.XMLHTTP");
				} catch (e) {}
			}
		}

		return xmlHttp;
	}

	function sendRequest(xml, tag) {
		if (isRequestPending || linkRecoveryActive) {
			// ignore if this is the same message as the request already in progress,
			// or if we're reconnecting and the same request has been made multiple times
			if (xml == currentRequest || (linkRecoveryActive && pendingRequests.length > 0 && xml == pendingRequests[pendingRequests.length - 1].xml)) {
				return;
			}

			// look for any existing requests with the same tag
			var add = true;

			if (tag !== undefined) {
				var len = pendingRequests.length;
				for (var i=0; i < len; i++) {
					if (pendingRequests[i].tag == tag) {
						pendingRequests[i].xml = xml;
						add = false;
						break;
					}
				}
			}

			if (add) {
				var req = {};
				req.xml = xml;
				req.tag = tag;

				pendingRequests.push(req);
			}

			return;
		}

		if (http_request && http_request.readyState != 4) {
			http_request.abort();
		}

		http_request = createHTTPRequest();

		isRequestPending = true;
		currentRequest = xml;

	    http_request.open('POST', serverURL + '/xml', true);
		http_request.onreadystatechange = processResponse;
	    http_request.send('xml=' + xml);
	}

	function sendRestRequest(url, func) {
		var request = createHTTPRequest();
	    request.open('GET', serverURL + url, true);
		request.onreadystatechange = function(evt) {
			try
			{
				if (request.readyState != 4) {
					return;
				}

				if (request.status != 200 || !request.responseText) {
					if (linkRecoveryActive) {
						if (linkRecoveryTimeout === undefined) {
							linkRecoveryTimeout = setTimeout(function() {
								linkRecoveryTimeout = undefined;
								getState(lastStateRev - 1);
							}, 1000);
						}
					} else {
						enterLinkRecovery();
					}
					return;
				}

				var response = JSON.parse(request.responseText);
				if (!response) {
					return;
				}

				try {
					func(response);
				} catch (e) {
					writeLog('Error processing response to ' + url, e);
				}
			} catch (e) {
				writeLog('', e);
				return;
			}
		};

	    request.send(null);
	}

	function getState(rev) {
		var url = '/state.json';
		if (rev !== undefined) {
			url += '?rev=' + rev;
		}

		lastStateRev = rev;
		sendRestRequest(url, processStateResponse);
	}

	function getSpeakers() {
		sendRestRequest('/speakers.json', function(res) {
			if (res.speakers) {
				self.speakers = res.speakers;
			} else {
				self.speakers = [];
			}

			if (self.onSpeakerListChanged) {
				self.onSpeakerListChanged(self.speakers);
			}
		});
	}

	function processResponse() {
		try
		{
			if (http_request.readyState != 4) {
				return;
			}

			currentRequest = '';
			isRequestPending = false;

			if (http_request.status != 200 || !http_request.responseXML) {
				return;
			}

			var root = http_request.responseXML.documentElement;
			if (!root) {
				return;
			}

			try {
				// We've received a response, if we're in link recovery request our new state
				if (linkRecoveryActive) {
					getState();
				}

				processResponseForErrors(root);
			} catch (e) {
				var msgName = "";
				if (root.tagName) {
					msgName = root.tagName;
				}

				writeLog('Error processing response ' + msgName, e);
			}

			// send the next pending request, if any
			if (pendingRequests.length > 0) {
				var req = pendingRequests.shift();
				sendRequest(req.xml, req.tag);
			}
		} catch (e) {
			writeLog('', e);
			return;
		}
	}

	function processResponseForErrors(xml) {
		var val_nodes = xml.getElementsByTagName('returnCode');
		if (val_nodes.length > 0)
		{
			if (val_nodes[0].firstChild && self.onResponseError)
			{
				self.onResponseError(getDataInt(val_nodes[0]));
			}
		}
	}

	function processStateResponse(state) {
		if (linkRecoveryActive) {
			// we've reconnected
			clearTimeout(linkRecoveryTimeout);
			linkRecoveryTimeout = undefined;
			linkRecoveryActive = false;
			if (pendingRequests.length > 0) {
				var req = pendingRequests.shift();
				sendRequest(req.xml, req.tag);
			}
		}

		var currentItem = self.currentMedia;
		var newItem = new MediaItem(state.currentItem);

		if (!currentItem.equals(newItem)) {
			self.currentMedia = newItem;

			if (self.onCurrentMediaChanged) {
				self.onCurrentMediaChanged(newItem);
			}

			if (self.onRatingChanged) {
				self.onRatingChanged(newItem.rating);
			}
		}

		var artURL = "";
		if (state.artworkURL) {
			artURL = serverURL + state.artworkURL + '&width=' + self.albumArtSize;
		}

		if (self.currentArtURL != artURL) {
			self.currentArtURL = artURL;
			if (self.onAlbumArtChanged) {
				self.onAlbumArtChanged(artURL);
			}
		}

		if (self.playingState != state.playingState) {
			self.playingState = state.playingState;
			if (self.onPlayingStateChanged) {
				self.onPlayingStateChanged(state.playingState);
			}
		}

		if (self.volume != state.volume) {
			self.volume = state.volume;
			if (self.onVolumeChanged) {
				self.onVolumeChanged(state.volume);
			}
		}

		if (self.shuffleState != state.shuffle) {
			self.shuffleState = state.shuffle;
			if (self.onShuffleStateChanged) {
				self.onShuffleStateChanged(self.shuffleState);
			}
		}

		var repeatState = (state.repeat != 0);
		if (self.repeatState != repeatState) {
			self.repeatState = repeatState;
			if (self.onRepeatStateChanged) {
				self.onRepeatStateChanged(self.repeatState);
			}
		}

		if (self.playlistPosition != state.playlistPosition) {
			self.playlistPosition = state.playlistPosition;
			if (self.onPlaylistPositionChanged) {
				self.onPlaylistPositionChanged(self.playlistPosition);
			}
		}

		// Always update the playback position to the server's current time
		self.playbackPosition = state.playbackPosition;
		if (self.onPlaybackPositionChanged) {
			self.onPlaybackPositionChanged(self.playbackPosition);
		}

		if (state.playlistUpdated === true) {
			if (self.onPlaylistChanged) {
				self.onPlaylistChanged();
			}
		}

		if (state.speakersUpdated === true) {
			getSpeakers();
		}

		if (state.libraryUpdated === true) {
			if (self.onMediaLibraryChanged) {
				self.onMediaLibraryChanged();
			}
		}

		getState(state.stateRevision);
	}

	function requestInitialState() {
		getState();

		if (self.onConnectedToPlayer) {
			self.onConnectedToPlayer();
		}
	}

	/* Monitors for pauses in the application's JavaScript execution,
	such as when the device enters sleep mode. When the device wakes,
	ping the server to ensure our link with it is still active. In some
	cases the phone will take a few seconds to restore Wi-Fi, during which
	time HTTP requests will be made through EDGE and take a long time to
	fail. */
	function sleepMonitor() {
		var currentTime = new Date();

		if ((currentTime - sleepMonitorTime > 7000) && (!linkRecoveryActive))
		{
			enterLinkRecovery();
		}

		sleepMonitorTime = currentTime;
	}

	function enterLinkRecovery() {
		if (linkRecoveryActive) {
			return;
		}

		// cancel any in-progress request
		if (http_request) {
			http_request.abort();
			http_request = null;
		}

		isRequestPending = false;
		linkRecoveryActive = true;

		getState(lastStateRev - 1);
	}

	function getData(elem) {
		if (elem.firstChild) {
			return escapeHtmlEntities(elem.firstChild.data);
		} else {
			return '';
		}
	}

	function getDataInt(elem) {
		if (elem.firstChild) {
			return parseInt(elem.firstChild.data);
		} else {
			return -1;
		}
	}

	function writeLog(msg, e) {
		if (debug) {
			throw e;
		}

		try {
			if (e) {
				msg += ': ' + e.toString();

				if (e.lineNumber) {
					msg += ' at line ' + e.lineNumber;
				} else if (e.line) {
					msg += ' at line ' + e.line;
				}

				if (e.fileName) {
					msg += ' of ' + e.fileName;
				}

				if (e.stack) {
					msg += '\nstack:\n' + e.stack;
				}
			}

			console.log(msg);
		} catch (loge) {}
	}
}

function escapeHtmlEntities(input) {
	return input.
	replace(/&/gmi, '&amp;').
	replace(/"/gmi, '&quot;').
	replace(/>/gmi, '&gt;').
	replace(/</gmi, '&lt;')
}
